@@ -1,5 +1,5 @@
 import logging
-from collections import namedtuple
+from collections import Counter, namedtuple
 from dataclasses import dataclass, field, fields
 from time import monotonic
 from typing import Iterable, Iterator, NamedTuple, Optional, Type, Union
@@ -22,6 +22,9 @@ from hcube.api.models.dimensions import (
     DateTimeDimension,
     Dimension,
     IntDimension,
+    IPv4Dimension,
+    IPv6Dimension,
+    MapDimension,
     StringDimension,
 )
 from hcube.api.models.filters import (
@@ -66,6 +69,12 @@ class TableMetaParams:
     # You can switch it off by using the following meta parameter with a False value.
     use_skip_indexes_if_final: bool = True
 
+    def use_sign_col(self):
+        return self.engine == "CollapsingMergeTree"
+
+    def use_final(self):
+        return self.engine.endswith("MergeTree") and self.engine != "MergeTree"
+
 
 class ClickhouseCubeBackend(CubeBackend):
 
@@ -80,6 +89,8 @@ class ClickhouseCubeBackend(CubeBackend):
         (DateTimeDimension, "DateTime64"),
         (FloatMetric, "Float"),
         (IntMetric, "Int"),
+        (IPv4Dimension, "IPv4"),
+        (IPv6Dimension, "IPv6"),
     )
     default_settings = {}
 
@@ -96,6 +107,7 @@ class ClickhouseCubeBackend(CubeBackend):
         self.client_attrs = client_attrs
         self.pool = ChPool(database=database, **client_attrs)
         self._table_exists = {}
+        self._query_counts = Counter()
 
     def initialize_storage(self, cube: Type[Cube]) -> None:
         self._create_table(cube)
@@ -118,10 +130,15 @@ class ClickhouseCubeBackend(CubeBackend):
             client.execute(f"USE {self.database}")
             meta = self.get_table_meta(cube)
             clean_records = cube.cleanup_records(records)
+            if meta.use_sign_col():
+                data = ({**rec._asdict(), meta.sign_col: 1} for rec in clean_records)
+            else:
+                data = ({**rec._asdict()} for rec in clean_records)
             client.execute(
                 f"INSERT INTO {self.database}.{self.cube_to_table_name(cube)} VALUES ",
-                ({**rec._asdict(), meta.sign_col: 1} for rec in clean_records),
+                data,
             )
+            self._query_counts[cube.__name__] += 1
 
     def get_records(
         self,
@@ -148,6 +165,7 @@ class ClickhouseCubeBackend(CubeBackend):
         with self.pool.get_client() as client:
             output = client.execute(text, params)
             logger.debug(f"Query time: {monotonic() - start: .3f} s")
+            self._query_counts[query.cube.__name__] += 1
             for rec in output:
                 yield result(*rec)
 
@@ -175,6 +193,8 @@ class ClickhouseCubeBackend(CubeBackend):
             )
 
         meta = self.get_table_meta(query.cube)
+        if meta.engine != "CollapsingMergeTree":
+            raise ConfigurationError("Delete query is only supported for collapsing merge trees")
         table = f"{self.database}.{self.cube_to_table_name(query.cube)}"
         where_parts = []
         params = {}
@@ -219,6 +239,7 @@ class ClickhouseCubeBackend(CubeBackend):
         start = monotonic()
         with self.pool.get_client() as client:
             client.execute(whole_text, params)
+            self._query_counts[query.cube.__name__] += 1
             logger.debug(f"Query time: {monotonic() - start: .3f} s")
 
     def delete_records_hard(self, query: CubeQuery) -> None:
@@ -251,6 +272,7 @@ class ClickhouseCubeBackend(CubeBackend):
         start = monotonic()
         with self.pool.get_client() as client:
             client.execute(text, params)
+            self._query_counts[query.cube.__name__] += 1
             logger.debug(f"Query time: {monotonic() - start: .3f} s")
 
     @classmethod
@@ -295,7 +317,7 @@ class ClickhouseCubeBackend(CubeBackend):
                     select_parts.append(grp.name)
             for agg in query.aggregations:
                 select_part, agg_params = self._translate_aggregation(
-                    agg, None if matview else meta.sign_col
+                    agg, None if (matview or not meta.use_sign_col()) else meta.sign_col
                 )
                 params.update(agg_params)
                 if agg.name not in fields:
@@ -311,7 +333,7 @@ class ClickhouseCubeBackend(CubeBackend):
             for metric in query.cube._metrics.values():
                 fields.append(metric.name)
                 select_parts.append(metric.name)
-            final = True  # there are no aggregations, we need to use FINAL
+            final = meta.use_final()  # there are no aggregations, it depends on engine
 
         select = ", ".join(select_parts) + append_to_select
         group_by = ", ".join(grp.name for grp in query.groups)
@@ -450,10 +472,13 @@ class ClickhouseCubeBackend(CubeBackend):
         """
         name = self.cube_to_table_name(cube)
         meta = self.get_table_meta(cube)
-        fields = [
-            f"{dim.name} {self._ch_type(dim)}"
-            for dim in list(cube._dimensions.values()) + list(cube._metrics.values())
-        ]
+        fields = []
+        for dim in list(cube._dimensions.values()) + list(cube._metrics.values()):
+            field = f"{dim.name} {self._ch_type(dim)}"
+            if ch_spec := dim.kwargs.get("clickhouse"):
+                if codec := ch_spec.get("compression_codec"):
+                    field += f" CODEC({codec}, LZ4)"
+            fields.append(field)
         field_part = ", ".join(fields)
         # indexes
         idx_part = ", ".join([idx.definition() for idx in meta.indexes])
@@ -468,7 +493,7 @@ class ClickhouseCubeBackend(CubeBackend):
                     f"Only cube dimensions may be part of the sorting key. These are extra: "
                     f"'{list(key_dim_names-cube_dim_names)}'"
                 )
-            if cube_dim_names - key_dim_names:
+            if meta.use_final() and cube_dim_names - key_dim_names:
                 logger.warning(
                     f"Dimensions '{list(cube_dim_names-key_dim_names)}' is missing from "
                     f"sorting_key, it will be collapsed in merge."
@@ -505,17 +530,26 @@ class ClickhouseCubeBackend(CubeBackend):
         )
         logger.debug(command)
         with self.pool.get_client() as client:
+            self._query_counts[cube.__name__] += 1
             client.execute(command)
 
     def _ch_type(self, dimension: Union[Dimension, Metric]) -> str:
         if isinstance(dimension, ArrayDimension):
             subtype = self._ch_type(dimension.dimension)
             return f"Array({subtype})"
+        if isinstance(dimension, MapDimension):
+            key_subtype = self._ch_type(dimension.key_dimension)
+            value_subtype = self._ch_type(dimension.value_dimension)
+            return f"Map({key_subtype}, {value_subtype})"
         for dim_cls, ch_type in self.dimension_type_map:
             if isinstance(dimension, (IntDimension, IntMetric)):
                 sign = "U" if not dimension.signed else ""
                 ch_type = f"{sign}{ch_type}{dimension.bits}"
             if isinstance(dimension, dim_cls):
+                # conversions specific for clickhouse
+                if ch_spec := dimension.kwargs.get("clickhouse"):
+                    if ch_spec.get("low_cardinality"):
+                        ch_type = f"LowCardinality({ch_type})"
                 if hasattr(dimension, "null") and dimension.null:
                     return f"Nullable({ch_type})"
                 return ch_type
@@ -550,6 +584,7 @@ class ClickhouseCubeBackend(CubeBackend):
         )
         logger.debug(command)
         with self.pool.get_client() as client:
+            self._query_counts[cube.__name__] += 1
             client.execute(command)
 
     def _create_projection(
@@ -560,7 +595,7 @@ class ClickhouseCubeBackend(CubeBackend):
         aggregs = [
             self._translate_aggregation(agg, meta.sign_col)[0] for agg in matview._aggregations
         ]
-        if matview.preserve_sign:
+        if matview.preserve_sign and meta.use_sign_col():
             aggregs.append(f"SUM({meta.sign_col}) AS _{meta.sign_col}")
         agg_part = ", ".join(aggregs)
         table_name = self.cube_to_table_name(cube)
@@ -572,10 +607,12 @@ class ClickhouseCubeBackend(CubeBackend):
         logger.debug(command)
         with self.pool.get_client() as client:
             client.execute(command)
+            self._query_counts[cube.__name__] += 1
             if populate:
                 client.execute(
                     f"ALTER TABLE {self.database}.{table_name} MATERIALIZE PROJECTION {view_name}"
                 )
+                self._query_counts[cube.__name__] += 1
 
     def _create_dictionaries(self, cube: Type[Cube]):
         meta = self.get_table_meta(cube)
@@ -586,21 +623,25 @@ class ClickhouseCubeBackend(CubeBackend):
                     f"SELECT comment FROM system.dictionaries "
                     f"WHERE database = '{self.database}' AND name = '{dict_def.name}'"
                 )
+                self._query_counts[cube.__name__] += 1
                 if out:
                     if out[0][0] != f"blake2:{dict_def.checksum}":
                         logger.info(
                             'Dictionary "%s" definition has changed, recreating', dict_def.name
                         )
                         client.execute(dict_def.drop_sql(database=self.database))
+                        self._query_counts[cube.__name__] += 1
                     else:
                         continue
                 client.execute(dict_def.definition_sql(database=self.database))
+                self._query_counts[cube.__name__] += 1
 
     def _drop_dictionaries(self, cube: Type[Cube]):
         meta = self.get_table_meta(cube)
         with self.pool.get_client() as client:
             for dict_def in meta.dictionaries:
                 client.execute(dict_def.drop_sql(database=self.database))
+                self._query_counts[cube.__name__] += 1
 
     def _translate_transform(self, transform: Transform) -> (str, str):
         """
@@ -629,3 +670,4 @@ class ClickhouseCubeBackend(CubeBackend):
             client.execute(
                 f"DROP TABLE IF EXISTS {self.database}.{self.cube_to_table_name(cube)} SYNC"
             )
+            self._query_counts[cube.__name__] += 1
