@@ -1,9 +1,11 @@
 import logging
 from collections import namedtuple
+from ipaddress import IPv4Address, IPv6Address
 from typing import Iterable, Iterator, NamedTuple, Optional, Type, Union
 
 import psycopg2
 from decouple import config
+from psycopg2.extras import Json
 
 from hcube.api.backend import CubeBackend
 from hcube.api.exceptions import ConfigurationError
@@ -15,6 +17,8 @@ from hcube.api.models.dimensions import (
     DateTimeDimension,
     Dimension,
     IntDimension,
+    IPDimension,
+    MapDimension,
     StringDimension,
 )
 from hcube.api.models.filters import (
@@ -68,6 +72,7 @@ class PostgresCubeBackend(CubeBackend):
         (DateTimeDimension, "Timestamp"),
         (FloatMetric, "Float"),
         (IntMetric, "Integer"),
+        (IPDimension, "Inet"),
     )
 
     def __init__(
@@ -94,10 +99,22 @@ class PostgresCubeBackend(CubeBackend):
         values_part = ", ".join(f"%({field})s" for field in cube.record_type()._fields)
         self._execute(
             f"INSERT INTO {self.schema}.{self.cube_to_table_name(cube)} VALUES ({values_part});",
-            [{**rec._asdict()} for rec in clean_records],
+            [{**self._to_sql_dict(rec)} for rec in clean_records],
             exec_many=True,
         )
 
+    @classmethod
+    def _to_sql_dict(cls, rec):
+        ret = {}
+        for field, value in rec._asdict().items():
+            if type(value) is dict:
+                ret[field] = Json(value)
+            elif isinstance(value, (IPv4Address, IPv6Address)):
+                ret[field] = str(value)
+            else:
+                ret[field] = value
+        return ret
+
     def get_records(self, query: CubeQuery) -> Iterator[NamedTuple]:
         text, params, fields = self._prepare_db_query(query)
         logger.debug('Query: "%s", params: "%s"', text, params)
@@ -292,6 +309,11 @@ class PostgresCubeBackend(CubeBackend):
             subtype, _subnull = self._ch_type_parts(dimension.dimension)
             # we ignore the null definition
             return f"{subtype}[]", ""
+        if isinstance(dimension, MapDimension):
+            keyb_subtype, _subnull = self._ch_type_parts(dimension.key_dimension)
+            value_subtype, _subnull = self._ch_type_parts(dimension.value_dimension)
+            # we ignore the null definition
+            return "jsonb", ""
         for dim_cls, ch_type in self.dimension_type_map:
             if isinstance(dimension, dim_cls):
                 if hasattr(dimension, "null") and dimension.null:
